/*!
 * © 2021 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals ROT, Result */
/* exported OpeningMove, Contination, Bout */
class OpeningMove {
	constructor(name) {
		this._name = name;
	}
	toString() {
		return this._name;
	}
	getOdds(theirMove,us,them) {
		throw new Error("getOdds not provided for "+this.toString()+".getOdds("+theirMove.toString()+","+us.toString()+","+them.toString()+")");
	}
}
OpeningMove.Charge = new OpeningMove("Charge");
OpeningMove.Charge.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 39; break;
		case theirMove === OpeningMove.Attack:
			odds = 65; break;
		case theirMove === OpeningMove.StepAside:
			odds = 20; break;
		case theirMove === OpeningMove.Slap:
			odds = 54; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.Charge");
	}
	odds += (us.getSpeed() - them.getWeight())/2.5;
	return odds;
};
OpeningMove.Charge.bothSameOpeningMoveNeedingRestart = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" charge.");
	return false;
};
OpeningMove.Charge.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("They crash heavily together in the middle of the dohyo.");
};
OpeningMove.Charge.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" charges "+them.toString()+" but "+them.toString()+" absorbs the charge.");
};
OpeningMove.Charge.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" charges "+them.toString()+" pushing "+them.toString()+" all the way to the edge of the dohyo and out. "+us.toString()+" wins by Oshi-dashi.");
};
OpeningMove.Attack = new OpeningMove("Attack");
OpeningMove.Attack.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 20; break;
		case theirMove === OpeningMove.Attack:
			odds = 39; break;
		case theirMove === OpeningMove.StepAside:
			odds = 65; break;
		case theirMove === OpeningMove.Slap:
			odds = 43; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.Attack");
	}
	odds += (us.getStrength() - them.getStrength())/2.5;
	return odds;
};
OpeningMove.Attack.bothSameOpeningMoveNeedingRestart = function(result,us,them) {
	result.addText("Both "+us.toString()+" and "+them.toString()+" go for the belt.");
	return false;
};
OpeningMove.Attack.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("They each get a grip on each other's belt.");
};
OpeningMove.Attack.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" grabs the belt of "+them.toString()+".");
};
OpeningMove.Attack.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" grabs the belt of "+them.toString()+" and pushes "+them.toString()+" all the way to the edge of the dohyo and out of it. "+us.toString()+" wins by Yori-kiri.");
};
OpeningMove.StepAside = new OpeningMove("Step Aside");
OpeningMove.StepAside.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 65; break;
		case theirMove === OpeningMove.Attack:
			odds = 20; break;
		case theirMove === OpeningMove.StepAside:
			odds = 39; break;
		case theirMove === OpeningMove.Slap:
			odds = 31; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.StepAside");
	}
	odds += (us.getAgility() - them.getSpeed())/2.5;
	return odds;
};
OpeningMove.StepAside.bothSameOpeningMoveNeedingRestart = function(_result,_us,_them) {
	return true;
};
OpeningMove.StepAside.addBothTriedMoveAndSucceededDescription = function(_result) {
	throw new Error("Both rikishi succeeded in a step aside move - which the code should have prevented");
};
OpeningMove.StepAside.addSuccessfulDescription = function(result,us,them,theirSuccessfulMove) {
	if (theirSuccessfulMove === OpeningMove.Charge) {
		result.addText(us.toString()+" steps to the side of "+them.toString()+".");
	} else {
		result.addText(us.toString()+" steps to the side and behind "+them.toString()+".");
	}
};
OpeningMove.StepAside.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" steps to the side and behind "+them.toString()+" pushing "+them.toString()+" out of the dohyo. "+us.toString()+" wins by Okuri-dashi.");
};
OpeningMove.Slap = new OpeningMove("Slap");
OpeningMove.Slap.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === OpeningMove.Charge:
			odds = 31; break;
		case theirMove === OpeningMove.Attack:
			odds = 43; break;
		case theirMove === OpeningMove.StepAside:
			odds = 54; break;
		case theirMove === OpeningMove.Slap:
			odds = 39; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for OpeningMove.Slap");
	}
	odds += (us.getStrength() - them.getWeight())/2.5;
	return odds;
};
OpeningMove.Slap.bothSameOpeningMoveNeedingRestart = function(_result,_us,_them) {
	return true;
};
OpeningMove.Slap.addBothTriedMoveAndSucceededDescription = function(_result) {
	throw new Error("Both rikishi succeeded in a slap move - which the code should have prevented");
};
OpeningMove.Slap.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" slaps "+them.toString()+".");
};
OpeningMove.Slap.addWinDescription = function(result,us,them,theirMove) {
	if (theirMove === OpeningMove.Charge) {
		result.addText(us.toString()+" slaps down "+them.toString()+" as they charge, winning by Hataki-komi.");
	} else {
		result.addText(us.toString()+" slaps down "+them.toString()+" winning by Tsuki-dashi.");
	}
};

class Continuation {
	constructor(name) {
		this._name = name;
	}
	toString() {
		return this._name;
	}
	getOdds(theirMove,us,them) {
		throw new Error("getOdds not provided for "+this.toString()+".getOdds("+theirMove.toString()+","+us.toString()+","+them.toString()+")");
	}
}
Continuation.Lever = new Continuation("Throw");
Continuation.Lever.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 39; break;
		case theirMove === Continuation.Lift:
			odds = 20; break;
		case theirMove === Continuation.Push:
			odds = 65; break;
		case theirMove === Continuation.Pull:
			odds = 54; break;
		case theirMove === Continuation.Trip:
			odds = 54; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Lever");
	}
	odds += (us.getAgility() - them.getStrength())/2.5;
	return odds;
};
Continuation.Lever.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to throw their opponent, but neither gains the advantage.");
};
Continuation.Lever.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" tries to throws "+them.toString()+".");
};
Continuation.Lever.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" throws "+them.toString()+". "+us.toString()+" wins by Uwate-nage.");
};
Continuation.Lift = new Continuation("Lift");
Continuation.Lift.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 65; break;
		case theirMove === Continuation.Lift:
			odds = 39; break;
		case theirMove === Continuation.Push:
			odds = 20; break;
		case theirMove === Continuation.Pull:
			odds = 43; break;
		case theirMove === Continuation.Trip:
			odds = 65; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Lift");
	}
	odds += (us.getStrength() - them.getWeight())/2.5;
	return odds;
};
Continuation.Lift.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to lift their opponent, but neither gains the advantage.");
};
Continuation.Lift.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" attempts to lift "+them.toString()+".");
};
Continuation.Lift.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is lifted out of the dohyo and "+us.toString()+" wins by Yori-taoshi.");
};
Continuation.Push = new Continuation("Push");
Continuation.Push.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 20; break;
		case theirMove === Continuation.Lift:
			odds = 65; break;
		case theirMove === Continuation.Push:
			odds = 39; break;
		case theirMove === Continuation.Pull:
			odds = 31; break;
		case theirMove === Continuation.Trip:
			odds = 31; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Push");
	}
	odds += (us.getWeight() - them.getWeight())/2.5;
	return odds;
};
Continuation.Push.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to push their opponent out of the dohyo, but neither gains the advantage.");
};
Continuation.Push.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" pushs "+them.toString()+", but "+them.toString()+" remains in the dohyo.");
};
Continuation.Push.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is pushed out of the dohyo and "+us.toString()+" wins by Oshi-dashi.");
};
Continuation.Pull = new Continuation("Pull");
Continuation.Pull.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 31; break;
		case theirMove === Continuation.Lift:
			odds = 43; break;
		case theirMove === Continuation.Push:
			odds = 54; break;
		case theirMove === Continuation.Pull:
			odds = 39; break;
		case theirMove === Continuation.Trip:
			odds = 39; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Pull");
	}
	odds += (us.getStrength() - them.getWeight())/2.5;
	return odds;
};
Continuation.Lever.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to pull their opponent to the ground, but neither gains the advantage.");
};
Continuation.Pull.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" pulls on "+them.toString()+".");
};
Continuation.Pull.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(them.toString()+" is pulled to the clay and "+us.toString()+" wins by Hiki-otoshi.");
};
Continuation.Trip = new Continuation("Trip");
Continuation.Trip.getOdds = function(theirMove,us,them) {
	let odds;
	switch(true) {
		case theirMove === Continuation.Lever:
			odds = 20; break;
		case theirMove === Continuation.Lift:
			odds = 65; break;
		case theirMove === Continuation.Push:
			odds = 65; break;
		case theirMove === Continuation.Pull:
			odds = 20; break;
		case theirMove === Continuation.Trip:
			odds = 39; break;
		default:
			throw new Error("Unknown countermove "+theirMove.toString()+" when determing odds for Continuation.Trip");
	}
	odds += (us.getSpeed() - them.getAgility())/2.5;
	return odds;
};
Continuation.Lever.addBothTriedMoveAndSucceededDescription = function(result) {
	result.addText("Both rikishi attempt to trip their opponent, but neither gains the advantage.");
};
Continuation.Trip.addSuccessfulDescription = function(result,us,them,_theirSuccessfulMove) {
	result.addText(us.toString()+" tries to trips "+them.toString()+".");
};
Continuation.Trip.addWinDescription = function(result,us,them,_theirMove) {
	result.addText(us.toString()+" trips "+them.toString()+" and "+us.toString()+" wins by Soto-gake.");
};

class Bout {
	static resolve(east,west) {
		const result = new Result();
		Bout._resolveMoves(true,result,east,west);
		while(result.getWinner() === null) {
			result.newline();
			Bout._resolveMoves(false,result,east,west);
		}
		return result;
	}
	static _resolveMoves(opening,result,east,west) {
		let eastmove,westmove;
		if (opening) {
			eastmove = east.getOpeningMove(west);
			westmove = west.getOpeningMove(east);
			while(eastmove === westmove && eastmove.bothSameOpeningMoveNeedingRestart(result,east,west)) {
				eastmove = east.getOpeningMove(west);
				westmove = west.getOpeningMove(east);
			}
		} else {
			eastmove = east.getContinuation(west);
			westmove = west.getContinuation(east);
		}

		const eastmovesuccess = (ROT.RNG.getPercentage() <= eastmove.getOdds(westmove,east,west));
		const westmovesuccess = (ROT.RNG.getPercentage() <= westmove.getOdds(eastmove,west,east));
		const eastsaves = east.saves();
		const westsaves = west.saves();
		const eastposswin = eastmovesuccess && !westsaves;
		const westposswin = westmovesuccess && !eastsaves;
		if (eastposswin && westposswin) {
			// Both rikishi succeeded and both failed to save - Need to determine who was the fastest
			const eastspeed = ROT.RNG.getPercentage()+east.getBalance()/2+ROT.RNG.getUniform()/5;
			const westspeed = ROT.RNG.getPercentage()+west.getBalance()/2+ROT.RNG.getUniform()/5;
			if (eastspeed > westspeed) {
				eastmove.addWinDescription(result,east,west,westmove);
				result.setWinner(east);
				result.setLoser(west);
			} else {
				westmove.addWinDescription(result,west,east,eastmove);
				result.setWinner(west);
				result.setLoser(east);
			}
		} else if (eastposswin) {
			eastmove.addWinDescription(result,east,west,westmove);
			result.setWinner(east);
			result.setLoser(west);
		} else if (westposswin) {
			westmove.addWinDescription(result,west,east,eastmove);
			result.setWinner(west);
			result.setLoser(east);
		} else if (eastmovesuccess && westmovesuccess) {
			// Both rikishi succeeded in their moves, but also both saved
			if (eastmove === westmove) {
				// Both made the same move
				eastmove.addBothTriedMoveAndSucceededDescription(result);
			} else {
				eastmove.addSuccessfulDescription(result,east,west,westmove);
				westmove.addSuccessfulDescription(result,west,east,eastmove);
			}
		} else if (eastmovesuccess) {
				eastmove.addSuccessfulDescription(result,east,west,null);
		} else 	if (westmovesuccess) {
				westmove.addSuccessfulDescription(result,west,east,null);
		} else {
			// Neither rikishi succeeded
			if (opening) {
				result.addText("Both rikishi crash together from the tachi-ai and lock up in the middle of the dohyo with neither having the advantage.");
			} else {
				result.addText("Both rikishi remain locked up in the middle of the dohyo with neither having the advantage.");
			}
		}
	}
}